#include <iostream>
#include <string>
#include <cstdlib>

#include <vtkVersion.h>
#include <vtkSmartPointer.h>
#include <vtkPolyData.h>
#include <vtkPolyDataMapper.h>
#include <vtkImageData.h>
#include <vtkPolyDataToImageStencil.h>
#include <vtkImageStencil.h>
#include <vtkPointData.h>
#include <vtkImageConstantPad.h>
#include <vtkMaskFields.h>
#include <vtkThreshold.h>
#include <vtkTransformFilter.h>
#include <vtkGeometryFilter.h>
#include <vtkTransform.h>
#include <vtkCellData.h>
#include <vtkRenderer.h>
#include <vtkRenderWindow.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkActor.h>
#include <vtkVoxelModeller.h>
#include <vtkLookupTable.h>
#include <vtkConeSource.h>
#include <vtkProperty.h>
#include <vtkDataSetMapper.h>
#include <vtkSTLReader.h>

#include <vtkCubeSource.h>
#include <vtkPolyData.h>
#include <vtkPoints.h>
#include <vtkGlyph3DMapper.h>
#include <vtkCellArray.h>


/**
 * This program reads a STL mesh and converts it into a volume 
 * representation (vtkImageData) where the foreground voxels are 1
 * and the background voxels are 0. The resulting image is visualized via 
 * vtk.
 **/
 
 
int main(int argc, char* argv[])
{
  // constants
  const int imageResolutionX = 50;
  const int imageResolutionY = 50;
  const int imageResolutionZ = 50;

  const unsigned char inval = 1;
  const unsigned char outval = 0;


  // vtkImageData for voxel representation storage  
  vtkSmartPointer<vtkImageData> voxelImage;

    
  // get commandline argument
  if(argc != 2)
  {
    std::string commandName(argv[0]);
    std::cout << "Usage: " << commandName << "<stl-filename>" << std::endl;
    std::cout << "or     " << commandName << "<stl-filename>" << std::endl;
    
    return EXIT_SUCCESS;
  }

  
  vtkNew<vtkSTLReader> reader;
  reader->SetFileName(argv[1]);
  reader->Update();

  vtkNew<vtkPolyData> pd;
  pd->DeepCopy(reader->GetOutput());
  
  // compute bounds for stl mesh polydata
  double bounds[6];
  pd->GetBounds(bounds);
  

  // use vtkPolyDataToImageStencil based on the example:
  // https://lorensen.github.io/VTKExamples/site/Cxx/PolyData/PolyDataToImageData/
  std::cout << "using vtkPolyDataToImageStencil for voxelization" << std::endl;

  vtkNew<vtkImageData> whiteImage;

  double spacing[3];                      // desired volume spacing
  spacing[0] = 1.0 / imageResolutionX;
  spacing[1] = 1.0 / imageResolutionY;
  spacing[2] = 1.0 / imageResolutionZ;

  whiteImage->SetSpacing(spacing);

  // compute dimensions
  int dim[3];
  for (int i = 0; i < 3; i++)
  {
      dim[i] = (int)(ceil((bounds[i * 2 + 1] - bounds[i * 2]) / spacing[i]));
  }
  whiteImage->SetDimensions(dim);
  whiteImage->SetExtent(0, dim[0] - 1, 0, dim[1] - 1, 0, dim[2] - 1);

  double origin[3];
  origin[0] = bounds[0] + spacing[0] / 2;
  origin[1] = bounds[2] + spacing[1] / 2;
  origin[2] = bounds[4] + spacing[2] / 2;
  whiteImage->SetOrigin(origin);
  whiteImage->AllocateScalars(VTK_UNSIGNED_CHAR,1);

  // fill the image with foreground voxels:
  vtkIdType count = whiteImage->GetNumberOfPoints();
  for (vtkIdType i = 0; i < count; ++i)
  {
      whiteImage->GetPointData()->GetScalars()->SetTuple1(i, inval);
  }

  // polygonal data --> image stencil:
  vtkNew<vtkPolyDataToImageStencil> pol2stenc;
  pol2stenc->SetInputData(pd);
  pol2stenc->SetOutputOrigin(origin);
  pol2stenc->SetOutputSpacing(spacing);
  pol2stenc->SetOutputWholeExtent(whiteImage->GetExtent());
  pol2stenc->Update();

  // cut the corresponding white image and set the background:
  vtkNew<vtkImageStencil> imgstenc;
  imgstenc->SetInputData(whiteImage);
  imgstenc->SetStencilConnection(pol2stenc->GetOutputPort());
  imgstenc->ReverseStencilOff();
  imgstenc->SetBackgroundValue(outval);
  imgstenc->Update();


  // Pad the volume so that we can change the point data into cell
  // data.
  voxelImage = imgstenc->GetOutput();

 
  // create point data for visualization via vtkGlyph3DMapper
  // based on the example code from 
  // https://www.vtk.org/Wiki/VTK/Examples/Cxx/Visualization/Glyph3DMapper 
  vtkNew<vtkPoints> points;

  for (int x = 0; x < dim[0]; x++)
  {
    for (int y = 0; y < dim[1]; y++)
    {
      for (int z = 0; z < dim[2]; z++)
      {
        char* pixel = static_cast<char*>(voxelImage->GetScalarPointer(x, y, z));
        
        if(pixel[0] == 1)
          points->InsertNextPoint(x, y, z);
      }
    }
  }

  vtkNew<vtkPolyData> polydata;
  polydata->SetPoints(points);

  vtkNew<vtkPolyData> glyph;

  vtkNew<vtkCubeSource> cubeSource;

  vtkNew<vtkGlyph3DMapper> glyph3Dmapper;
  glyph3Dmapper->SetSourceConnection(cubeSource->GetOutputPort());
  glyph3Dmapper->SetInputData(polydata);
  glyph3Dmapper->Update();
  

  vtkNew<vtkActor> actor;
  actor->SetMapper(glyph3Dmapper);
  actor->GetProperty()->SetColor(0.0, 1.0, 0.0);
  actor->GetProperty()->SetOpacity(0.5); 


  vtkNew<vtkRenderer> renderer;
  vtkNew<vtkRenderWindow> renderWindow;
  renderWindow->AddRenderer(renderer);
  
  vtkNew<vtkRenderWindowInteractor> renderWindowInteractor;
  renderWindowInteractor->SetRenderWindow(renderWindow);

  renderer->AddActor(actor);
  renderer->SetBackground(1.0, 1.0, 1.0);

  renderWindow->Render();
  renderWindowInteractor->Start();

  return EXIT_SUCCESS;
}

